// Fl_Shared_Image_Display - a widget to display FLTK's images
// (c) 2005 Nathan Vander Wilt.
//
// This program is licensed under the same terms as the FLTK 
// license that should have been included with this file. Programs
// must identify the use of this work.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// FLTK License for more details.
//

#ifndef Fl_Shared_Image_Display_H
#define Fl_Shared_Image_Display_H

#include <FL/Fl.H>
#include <FL/Fl_Widget.H>
#include <FL/Fl_Image.H>
#include <FL/fl_draw.H>
#include <FL/x.H>                 //needed for fl_parse_color
#include <stdio.h>

// Known issues:
//potential problems when XPM uses color names (from  from fl_parse_color on Win/OSX  and from assuming names don't have spaces)
//problems when pixmap bpp>sizeof(int)

//speed -
//already vastly improved from original version
//  use Bresenham/or Xiaolin Wu's line algorithm ideas? - don't integrate well with general transform
//  make more efficient get_pixel? (move range checking to line fcn?)
//  anything else?

class Fl_Shared_Image_Display : public Fl_Widget {
 int img_w;
 int img_h;

 union pixel {
  uchar rgb[3];
  struct {
   uchar r;
   uchar g;
   uchar b;
  };
 };

 union pixel_t {
  uchar rgb[4];
  struct {
   uchar r;
   uchar g;
   uchar b;
   uchar a;
  };
 };

 struct {
  bool set;
  int start;
  int bppix;
  pixel* palette;
  const char* const* img_line;
 } pixmp_;
 int xxmult;
 int xymult;
 int yxmult;
 int yymult;
 int xadd;
 int yadd;
 int rotate_;
 bool fill_;
 Fl_Color bgin;
 pixel bg_;

 pixel (*pxl_cb_)(void*,int,int);
 Fl_Image* img_;
 static void img_line(void* disp_widget, int x, int y, int d, uchar* data) { ((Fl_Shared_Image_Display*)disp_widget)->img_line_(x,y,d,data); }
 void img_line_(int, int, int, uchar*);
 void img_line_pre();
 static void img_bg_line(void* disp_widget, int x, int y, int d, uchar* data) { ((Fl_Shared_Image_Display*)disp_widget)->img_bg_line_(x,y,d,data); }
 void img_bg_line_(int, int, int, uchar*);

 static pixel get_pixel3(void* d,int x, int y) { return ((Fl_Shared_Image_Display*)d)->get_pixel3_(x,y); }
 pixel get_pixel3_(int,int);
 static pixel get_pixel1(void* d,int x, int y) { return ((Fl_Shared_Image_Display*)d)->get_pixel1_(x,y); }
 pixel get_pixel1_(int,int);
 static pixel get_pixel2(void* d,int x, int y) { return ((Fl_Shared_Image_Display*)d)->get_pixel2_(x,y); }
 pixel get_pixel2_(int,int);
 static pixel get_pixel4(void* d,int x, int y) { return ((Fl_Shared_Image_Display*)d)->get_pixel4_(x,y); }
 pixel get_pixel4_(int,int);

 static pixel get_bit_pixel(void* d,int x, int y) { return ((Fl_Shared_Image_Display*)d)->get_bit_pixel_(x,y); }
 pixel get_bit_pixel_(int,int);
 static pixel get_pix_pixel(void* d,int x, int y) { return ((Fl_Shared_Image_Display*)d)->get_pix_pixel_(x,y); }
 pixel get_pix_pixel_(int,int);

 pixel blend_pix_bg(pixel_t);
 pixel blend_lum_bg(uchar,uchar);

 void set_pxmp();
 uchar bit(int);
 inline int abs(int x) { return ( x > 0 ? x : -x ); }
 inline int intexp(int x, unsigned int n) { int r=1; for (int i=0; i<n; i++) r*=x; return r; } //return x to the n
 inline bool whitespace(const char* x) { if ( *x==' ' || *x=='\t' ) return true; else return false; }
protected:
 void draw();
public:
 void bgcolor(Fl_Color c) { bgin=c; pixmp_.set=false; redraw(); }
 void set_bgs() { Fl::get_color(bgin,bg_.r,bg_.g,bg_.b); }
 Fl_Color bgcolor() { return bgin; }
 void fill(int x) { if ( (x && fill_) || (!x && !fill_) ) return; if (x) fill_=true; else fill_=false; redraw(); }
 int fill() { return fill_; }
 void rotate(int x) { if (x==rotate_) return; rotate_=abs(x); if (rotate_ > 8 ) rotate_=rotate_%8; if (!rotate_) rotate_=1; redraw(); }
 int rotate() { return rotate_; }
 void image(Fl_Image* img) { img_=img; pixmp_.set=false; if (pixmp_.palette) delete [] pixmp_.palette; pixmp_.palette=0; redraw(); }
 Fl_Shared_Image_Display(int x, int y, int w, int h, const char *label=0) : Fl_Widget(x,y,w,h,label) {
  xxmult=1;
  xymult=0;
  yxmult=0;
  yymult=1;
  xadd=0;
  yadd=0;
  rotate_=1;
  fill_=false;
  bgin=FL_GRAY;
  img_=0;
  pixmp_.set=false;   //!!change to function so can delete pixmp_.palette if necessary
  pixmp_.palette=0;
 }
};

#endif //Fl_Share_Image_Display_H

