// Fl_Shared_Image_Display - a widget to display FLTK's images
// (c) 2005 Nathan Vander Wilt.
//
// This program is licensed under the same terms as the FLTK 
// license that should have been included with this file. Programs
// must identify the use of this work.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// FLTK License for more details.
//

#include "Fl_Shared_Image_Display.H"

void Fl_Shared_Image_Display::draw() {
 draw_box();

 int drw_x, drw_y, drw_w, drw_h;
 drw_x = x() + Fl::box_dx(box());
 drw_y = y() + Fl::box_dy(box());
 drw_w = w() - Fl::box_dw(box());
 drw_h = h() - Fl::box_dh(box());

 fl_push_clip(drw_x,drw_y,drw_w,drw_h);
 if (img_) {
  pxl_cb_=0;
  if ( img_->d() == 0 ) {                       //check for bitmaps
   img_w = img_->w();
   img_h = img_->h();
   pxl_cb_=get_bit_pixel;
  }
  else if ( img_->count() > 1 ) {              //check for pixmaps
   if ( !pixmp_.set ) set_pxmp();
   pxl_cb_=get_pix_pixel;
  }
  else {
   img_w = img_->w();
   img_h = img_->h();
   switch (img_->d()) {
    case 1: pxl_cb_=get_pixel1; break;
    case 2: pxl_cb_=get_pixel2; break;
    case 3: pxl_cb_=get_pixel3; break;
    case 4: pxl_cb_=get_pixel4; break;
   }
  }
  img_line_pre();
  if (pxl_cb_) fl_draw_image(img_line, this, drw_x, drw_y, drw_w, drw_h, 3);
 }
 else fl_draw_image(img_bg_line, this, drw_x, drw_y, drw_w, drw_h, 3);

 draw_label();
 fl_pop_clip();
}

void Fl_Shared_Image_Display::img_line_pre() {
 set_bgs();
 int drw_w = w() - Fl::box_dw(box());
 int drw_h = h() - Fl::box_dh(box());
 int xxs=0;
 int xys=0;
 int yys=0;
 int yxs=0;
 if ( !drw_w || !drw_h ) {                // save time and avoid div/0
  xxmult=0; xymult=0; yymult=0; yxmult=0;
  xadd=0; yadd=0;
  return;
 }

 switch (rotate_) {
  case 1: xxs=1; yys=1; break;
  case 2: xxs=-1; yys=1; break;
  case 3: xxs=-1; yys=-1; break;
  case 4: xxs=1; yys=-1; break;
  case 5: xys=1; yxs=1; break;
  case 6: xys=-1; yxs=1; break;
  case 7: xys=-1; yxs=-1; break;
  case 8: xys=1; yxs=-1; break;
 }

 if ( rotate_ > 4 ) {              //for orientations > 4, need swapped transform matrix
  xymult = 256*img_w/drw_h;
  yxmult = 256*img_h/drw_w;
  if (!fill_ && xymult>yxmult ) yxmult=xymult;
  if (!fill_ && yxmult>xymult ) xymult=yxmult;
 }
 else {
  xxmult = 256*img_w/drw_w;
  yymult = 256*img_h/drw_h;
  if (!fill_ && xxmult>yymult ) yymult=xxmult;
  if (!fill_ && yymult>xxmult ) xxmult=yymult;
 }

 xxmult*=xxs;
 xymult*=xys;
 yymult*=yys;
 yxmult*=yxs;

 xadd=(256*img_w-(drw_w-1)*xxmult-(drw_h-1)*xymult)/2;        // does this chop off outer edge? -nein
 yadd=(256*img_h-(drw_h-1)*yymult-(drw_w-1)*yxmult)/2;

}

void Fl_Shared_Image_Display::img_bg_line_(int x, int y, int w, uchar* data) {
 pixel* pdata=(pixel*)data;

 for ( int i=0; i<w; i++ ) {
  pdata[i] = bg_;
 }
}

void Fl_Shared_Image_Display::img_line_(int x, int y, int w, uchar* data) {
 int d= img_->d();
 pixel* pdata=(pixel*)data;

 int xadd_=y*xymult+xadd;
 int yadd_=y*yymult+yadd;
 int x_;
 int y_;


 if ( xxmult && !yxmult ) {
  y_=yadd_ >> 8;
  for ( int i=0; i<w; i++ ) {
   x_=((i+x)*xxmult+xadd_) >> 8;
   pdata[i] = (*pxl_cb_)(this,x_,y_);
  }
 }
 else if ( yxmult && !xxmult ) {
  x_=xadd_ >> 8;
  for ( int i=0; i<w; i++ ) {
   y_=((i+x)*yxmult+yadd_) >> 8;
   pdata[i] = (*pxl_cb_)(this,x_,y_);
  }
 }
 else {
  y_=yadd_ >> 8;
  for ( int i=0; i<w; i++ ) {
   x_=((i+x)*xxmult+xadd_) >> 8;
   pdata[i] = (*pxl_cb_)(this,x_,y_);
  }
 }
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::get_pixel3_(int x, int y) {   //get pixel from image, using bg_ for transparent/out of range
 int w=img_->w();
 
 if ( x >= w || x < 0 ) return bg_;
 if ( y >= img_->h() || y < 0 ) return bg_;
 
 return *((pixel*)img_->data()[0]+(y*w+x));
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::get_pixel1_(int x, int y) {   //get pixel from image, using bg_ for transparent/out of range
 int w=img_->w();
 pixel pxl;

 if ( x >= w || x < 0 ) return bg_;
 if ( y >= img_->h() || y < 0 ) return bg_;
 
 pxl.r=*((uchar*)img_->data()[0]+(y*w+x));
 pxl.g=*((uchar*)img_->data()[0]+(y*w+x));
 pxl.b=*((uchar*)img_->data()[0]+(y*w+x));

 return pxl;
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::get_pixel2_(int x, int y) {   //get pixel from image, using bg_ for transparent/out of range
 int w=img_->w();

 if ( x >= w || x < 0 ) return bg_;
 if ( y >= img_->h() || y < 0 ) return bg_;
 
 return blend_lum_bg( *((uchar*)img_->data()[0]+(y*w+x)*2), *((uchar*)img_->data()[0]+(y*w+x)*2+1) );
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::get_pixel4_(int x, int y) {   //get pixel from image, using bg_ for transparent/out of range
 int w=img_->w();

 if ( x >= w || x < 0 ) return bg_;
 if ( y >= img_->h() || y < 0 ) return bg_;
 
 return blend_pix_bg( *((pixel_t*)img_->data()[0]+(y*w+x)) );
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::blend_lum_bg(uchar lumin, uchar amount) {
 pixel blend;
 blend.r=bg_.r+amount*(lumin-bg_.r)/255;
 blend.g=bg_.g+amount*(lumin-bg_.g)/255;
 blend.b=bg_.b+amount*(lumin-bg_.b)/255;
 return blend;
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::blend_pix_bg(pixel_t color) {
 pixel blend;
 blend.r=bg_.r+color.a*(color.r-bg_.r)/255;
 blend.g=bg_.g+color.a*(color.g-bg_.g)/255;
 blend.b=bg_.b+color.a*(color.b-bg_.b)/255;
 return blend;
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::get_bit_pixel_(int x, int y) {   //get pixel from bitmapped image
 int w=img_->w();
 pixel pix;

 if ( x >= w || x < 0 ) return bg_;
 if ( y >= img_->h() || y < 0 ) return bg_;
 w += w%8 ? 8-w%8 : 0;                    //!!this does something good, why?

 pix.r= ( *((uchar*)img_->data()[0]+(y*w+x)/8) & bit(y*w+x) ) ? 0 : 255;
 pix.g= pix.r;
 pix.b= pix.r;
 return pix;
}

uchar Fl_Shared_Image_Display::bit(int x) {
 switch(x%8) {
  case 0: return 1;
  case 1: return 2;
  case 2: return 4;
  case 3: return 8;
  case 4: return 16;
  case 5: return 32;
  case 6: return 64;
  case 7: return 128;
 }
}

Fl_Shared_Image_Display::pixel Fl_Shared_Image_Display::get_pix_pixel_(int x, int y) {   //get pixel from pixmap image
 int w=img_w;

 if ( x >= w || x < 0 ) return bg_;
 if ( y >= img_h || y < 0 ) return bg_;

 int lookup;
 for (int i=0; i<pixmp_.bppix; i++) {
  if (!i) lookup=*((uchar*)pixmp_.img_line[pixmp_.start+y]+pixmp_.bppix*x);     // !! a tiny stab at optimizing for single bpp images
  else {
   int temp=*((uchar*)pixmp_.img_line[pixmp_.start+y]+pixmp_.bppix*x+i);
   lookup+=intexp(2,8*i)*temp;
  }
 }
 
 return pixmp_.palette[lookup];
}




void Fl_Shared_Image_Display::set_pxmp() {
 //Thanks to Ian MacArthur for a good example (flbundle.cxx) of how to do this!
 pixmp_.img_line=img_->data();
 int colours;

 sscanf(pixmp_.img_line[0],"%d %d %d %d", &img_w, &img_h, &colours, &pixmp_.bppix);
 pixmp_.start=img_->count()-img_h;

 if (pixmp_.palette) delete [] pixmp_.palette;
 pixmp_.palette=new pixel[intexp(2,8*pixmp_.bppix)];  //size is often 256 (pixmp_.bppix=1)

 if (colours<0) {              // binary palette
  uchar entry;

  for (int i=0; i < -colours; i++) {
    entry=pixmp_.img_line[1][i*4];
    pixmp_.palette[entry].r=pixmp_.img_line[1][i*4+1];
    pixmp_.palette[entry].g=pixmp_.img_line[1][i*4+2];
    pixmp_.palette[entry].b=pixmp_.img_line[1][i*4+3];
  }
  entry=pixmp_.img_line[1][0];
  if (entry==0x20) pixmp_.palette[entry]=bg_;        // set GIF transparent pixel
 }
 else {            //ASCII palette (assuming always XPM format, with the color we want set as 'c' color )
  unsigned int entry;
  pixel pxl;
  for (int i=1; i < colours+1; i++) {
   entry=0;
   for (int j=0; j < pixmp_.bppix; j++ ) entry+=intexp(2,8*j)*pixmp_.img_line[i][j];

   const char* ptr=pixmp_.img_line[i]+pixmp_.bppix+1;
   const char* nxt=0;
   char* buff=0;

   while (true) {
    ptr=strchr(ptr, 'c');
    if ( whitespace(ptr-1) && whitespace(ptr+1) ) {
     while ( whitespace(++ptr) ) if (*ptr=='\0') break;      // move through whitespace after the ' c ' character.
     nxt=(char*)ptr;
     while ( *nxt !='\0' ) if (whitespace(nxt++)) break;
     buff=new char[1+nxt-ptr];
     strncpy(buff,ptr,nxt-ptr);
     buff[nxt-ptr]='\0';

     if ( !strcmp(buff,"None") ) {                      //use our bgcolor for transparent rather than what fl_parse_color grabs
      pxl=bg_;
      break;
     }
     else if ( !fl_parse_color(buff, pxl.r, pxl.g, pxl.b) ) { /* printf("@%i - %s\n",i,buff); fflush(stdout); */ pxl=bg_; break; }       // Undocumented, from Fl_get_system_colors.cxx via <FL/x.H> !! This is broken on Windows/OSX if a color name is found rather than a #value
     else break;
    }
   }
   if (buff) delete [] buff;
   pixmp_.palette[entry]=pxl;
  }  
 }
 pixmp_.set=true;
}


